/* minesweeper for 3rd ed. Plan9
  * 
  * most of the ideas come from Charles Forsyth's 2nd ed
  * minesweeper. the bad code is entirely my fault.
  * I've changed the GUI to use control(2) and messed much 
  * with the clean data structs in the original code.
  * 
  * problems: button2/3 of the mouse cannot be used...
  *
  * andrey mirtchovski (andrey@lanl.gov)
  */

/* change this for different difficulty levels */
#define GET_NMINES(x, y)	(8 + ((x) + (y))/4)


#include <u.h>
#include <libc.h>
#include <draw.h>
#include <thread.h>
#include <keyboard.h>
#include <mouse.h>
#include <control.h>

typedef struct Cell Cell;

struct Cell {
	uchar	type;
	uchar	guess;
	uchar	near;
};

enum {
	Invalid = 0,
	None,
	Empty,
	Mined,
	Query	/* this is unused */
};

enum {
	Maxcount = 8,	/* max neighbours of a cell */
	Cellsize = 21,	/* each cell is Cellsize² pixels */
	MaxGrid = 30,	/* what is the maximum grid size -- 30x30
				  * this should be dynamically allocated
				  */
	testing = 0,
};

char *bstr[Maxcount+1] = {
	"b0",
	"b1",
	"b2",
	"b3",
	"b4",
	"b5",
	"b6",
	"b7",
	"b8",
};

char tmpstr[128];

//Control ***buttons; /* dinamically allocated buttons would be a better idea */
Cell 			cells[MaxGrid][MaxGrid];		/* probably best to add 'buttons' 
									  * to the Cell struct 
									  */
Control 		*buttons[MaxGrid][MaxGrid];
Controlset 	*cs  		= nil;
Channel 		*c 		= nil;		/* button events */
int 			dx, dy;
int 			nmines;
int			total;		
int 			ctldeletequits 	= 1;

static		Image	*bomb;
static		Image	*unknown;
static		Image	*query;
static		Image	*marked;
static		Image	*badmark;
static		Image	*digit[Maxcount+1];

static	Point	neighbours[] = {
	{-1,-1},{-1,0},{-1,1},
	{1,-1},{1,0},{1,1},
	{0,-1},{0,1},
	{0,0}
};
enum {
	Etime = 4,
};


/* this should be part of the library, methinks */
int
isactive(Control *a) 
{
	Control *c;
	
	for(c = a->controlset->actives; c; c = c->nextactive)
		if(c == a)
			return 1;
	
	return 0;
}


/* opens up the map.. if success is non-zero then the mines are marked with flags */
void
revealmap(int success)
{
	int i, j;

	/* clear the entire field and deactivate all buttons
 	  * someday the buttons will be dinamically allocated
	  * and this isactive()/deactivate() business will be 
 	  * removed
	  */
	for(i = 0; i < MaxGrid; i++) 
		for(j = 0; j < MaxGrid; j++) {
			if(success && cells[i][j].type == Mined)
				printctl(buttons[i][j]->ctl, "light marked");
			printctl(buttons[i][j]->ctl, "value 1");
		}
}



/* handles recursively the opening of the white spaces on the screen */
void
do_neighbours(int x, int y) 
{
	Point *p;
	
	if(--total == 0)
		revealmap(1);
	printctl(buttons[x][y]->ctl, "value 1");
	cells[x][y].guess = Empty;

	if(cells[x][y].near != 0)
		return;

	for(p=neighbours; p->x||p->y; p++)
		if(cells[x+p->x][y+p->y].guess == None) 
			do_neighbours(x+p->x, y+p->y);
}


static void
minefield(int n)
{
	int i, x, y;
	int tmp, tmp1;
	Point *p;


	for(i=0; i<n; i++) {
		do {
			tmp = nrand(dx-1);
			tmp1 = nrand(dy-1);
		} while(cells[tmp][tmp1].type==Mined);

		cells[tmp][tmp1].type = Mined;
		printctl(buttons[tmp][tmp1]->ctl, "light bomb");
		for(p=neighbours; p->x||p->y; p++) {
			cells[tmp+p->x][tmp1+p->y].near++;
		}
	}

	for(x=0; x<dx; x++)
		for(y=0; y<dy; y++) {
			if(cells[x][y].type != Mined) {
				printctl(buttons[x][y]->ctl, "light %q", bstr[cells[x][y].near]);	
				cells[x][y].type = None;
			}
			cells[x][y].guess = None;
		}

}


void
closemap(void)
{
	int i, j;

	/* clear the entire field and deactivate all buttons
 	  * someday the buttons will be dinamically allocated
	  * and this isactive()/deactivate() business will be 
 	  * removed
	  */
	for(i = 0; i < MaxGrid; i++) {
		for(j = 0; j < MaxGrid; j++) {
			if(isactive(buttons[i][j]))
				deactivate(buttons[i][j]);
			cells[i][j].type = None;
			cells[i][j].guess = 0;
			cells[i][j].near = 0;
		}
	}
}

void
resizecontrolset(Controlset*) 
{
	Rectangle r, r1;
	int i, j;

	if(getwindow(display, Refnone) < 0) 
		ctlerror("resized: %r");


	draw(screen, screen->r, display->black, nil, ZP);

	r = insetrect(screen->r, 10);
	if((dx = Dx(r)/Cellsize) >= MaxGrid)
		dx = MaxGrid;
	if((dy = Dy(r)/Cellsize) >= MaxGrid)
		dy = MaxGrid;

	closemap();
	nmines = GET_NMINES(dx, dy);
	total = dx*dy - nmines;
	minefield(nmines);


	r1 = Rect(r.min.x, r.min.y, r.min.x + Cellsize, r.min.y + Cellsize);
	for(j = 0; j < dy; j++, r1.min.y += Cellsize, r1.max.y += Cellsize) {
		for(i = 0; i < dx; i++, r1.min.x += Cellsize, r1.max.x += Cellsize) {
			printctl(buttons[i][j]->ctl, "image unknown");
			printctl(buttons[i][j]->ctl, "rect %R\nshow", r1);
			printctl(buttons[i][j]->ctl, "value 0");
			if(!isactive(buttons[i][j]))
				activate(buttons[i][j]);
		}
		r1.min.x = r.min.x;
		r1.max.x = r.min.x + Cellsize;
	}

}

void
ctlsetinit(void) 
{
	int i, j;

	c 	= chancreate(sizeof(char *), 0);
	cs 	= newcontrolset(screen, nil, nil, nil);


	for(j = 0; j < MaxGrid; j++) {
		for(i = 0; i < MaxGrid; i++) {
			sprint(tmpstr, "%d-%d", i, j);
			buttons[i][j]= createbutton(cs, tmpstr);
			controlwire(buttons[i][j], "event", c);
		}
	}
}


int 
loadimages(void)
{
	int fd, i;

	/* some more error checking may be a good idea... */
	if((fd = open("bitmaps/bomb.b", OREAD)) < 0) 
		return -1;
	bomb = readimage(display, fd, 0);
	close(fd);
	namectlimage(bomb, "bomb");

	if((fd = open("bitmaps/unknown.b", OREAD)) < 0) 
		return -1;
	unknown = readimage(display, fd, 0);
	close(fd);
	namectlimage(unknown, "unknown");

	if((fd = open("bitmaps/marked.b", OREAD)) < 0) 
		return -1;
	marked = readimage(display, fd, 0);
	close(fd);
	namectlimage(marked, "marked");

	if((fd = open("bitmaps/question.b", OREAD)) < 0) 
		return -1;
	query = readimage(display, fd, 0);
	close(fd);
	namectlimage(query, "query");

	if((fd = open("bitmaps/badmark.b", OREAD)) < 0) 
		return -1;
	badmark = readimage(display, fd, 0);
	close(fd);
	namectlimage(badmark, "badmark");

	for(i=0; i<=Maxcount; i++) {
		sprint(tmpstr, "bitmaps/b%d.b", i);
		if((fd = open(tmpstr, OREAD)) < 0) 
			return -1;
		digit[i] = readimage(display, fd, 0);
		close(fd);
		namectlimage(digit[i], bstr[i]);
	}
	return 0;
}
	
void
threadmain(int, char**)
{
	char *args[3];
	char *s;
	int n;

	srand(getpid()^time(0));


	initdraw(0, 0, "minesweeper");
	initcontrols();

	if(loadimages() < 0) {
		perror("cannot read images from bitmap/: %r");
		exits("images library not found");
	}

	ctlsetinit();
	resizecontrolset(cs);

	for(;;){
		
		s = recvp(c);
		n = tokenize(s, args, nelem(args));
		if(n==3 && strcmp(args[1], "value")==0){
			int x, y;
			char *tmp;
			
			/* this is crap */
			tmp = strchr(args[0], '-');
			*tmp++ = '\0';
			x = atoi(args[0]);
			y = atoi(tmp);
			
			if(cells[x][y].type == Mined) {
				revealmap(0);
			}
			else {
				printctl(buttons[x][y]->ctl, "value 1");
				cells[x][y].guess = Empty;
				if(cells[x][y].near == 0) {
					do_neighbours(x, y);
				}
				else if(--total == 0) {
					revealmap(1);
				}

			}
		}		
	}
	threadexitsall(nil);
}

